"""
Classes that parse Label Studio exports.

Does not know about Chart Review concepts like config files.
Simply parses and exposes Label Studio concepts.

Some of this parsing is a little generous (like allowing missing id, origin, type, or from_name
fields), mostly to accommodate tests without fully detailed input.
(Which, I think is fine - generous parsing is good anyway, and tests with fully specified exports
might make it harder to understand what the test is focusing on.)
"""

import dataclasses
import os
import sys

from chart_review import common, defines, errors


@dataclasses.dataclass(kw_only=True)
class Mention:
    """A piece of text and labels, very similar to LabeledText."""

    id: str
    text: str
    labels: defines.LabelSet
    from_name: str

    @staticmethod
    def parse(entry: dict) -> "Mention":
        # Check where we're going to find the labels/tags
        match entry.get("type", "labels").casefold():
            case "choices":
                # Looks like:
                # "value": {
                #   "text": "patient is deceased",
                #   "choices": ["False"]
                # },
                field = "choices"
                is_list = True
            case "datetime":
                # Looks like:
                # "value": {
                #   "text": "Nov 15",
                #   "datetime": "2018-11-15",
                # },
                field = "datetime"
                is_list = False
            case "labels":
                # Looks like:
                # "value": {
                #   "text": "patient has infection",
                #   "labels": ["Infection"]
                # },
                field = "labels"
                is_list = True
            case "textarea":
                # Looks like:
                # "value": {
                #   "text": ["free form"],
                # },
                field = "text"
                is_list = True
            case _:
                raise ValueError(f"Unrecognized Label Studio result type '{entry.get('type')}'.")

        value = entry.get("value", {})
        text = value.get("text", "") if field != "text" else ""
        field_value = value.get(field)
        if not field_value:
            labels = set()
        elif is_list:
            labels = {defines.Label(x) for x in field_value}
        else:
            labels = {defines.Label(field_value)}
        return Mention(
            id=entry.get("id", ""), text=text, labels=labels, from_name=entry.get("from_name", "")
        )


@dataclasses.dataclass(kw_only=True)
class Annotation:
    """All of a single source's mentions"""

    author: int
    mentions: list[Mention] = dataclasses.field(default_factory=list)

    @staticmethod
    def parse(entry: dict, data_keys: set[str]) -> "Annotation | None":
        author = entry.get("completed_by")
        if author is None:
            return None  # we don't know who annotated this!

        # Labels can be nested - e.g. there might be a toplevel label "Illness" and a sublabel
        # (maybe called "Illness Confirmed?") with a three-way choice of "confirmed", "suspected",
        # and "none of the above".
        # The way that would shows up as a mention in an export is a little odd (to my mind):
        # {
        #   "value": {
        #     "text": "text from note",
        #     "labels": ["Illness"]
        #   },
        #   "id": "HI1y_tlNwu",
        #   "from_name": "label",
        #   "to_name": "text",
        #   "type": "labels",
        # },
        # {
        #   "value": {
        #     "text": "text from note",
        #     "choices": ["confirmed"]
        #   },
        #   "id": "HI1y_tlNwu",
        #   "from_name": "Illness Confirmed?",
        #   "to_name": "text",
        #   "type": "choices",
        # },
        #
        # So you can see there that the "id" field is re-used, and the sublabel refers to the
        # parent via "from_name".

        # When parsing here, we'll first look for the toplevel entries (the first use of any given
        # "id" value). Then do a second pass for any sublabels and adjust the parent with the extra
        # info.
        mentions: list[Mention] = []
        toplevels: dict[str, Mention] = {}
        sublabels: list[Mention] = []
        for result in entry.get("result", []):
            mention = Mention.parse(result)
            if not mention.id or mention.id not in toplevels:
                # This is a toplevel mention
                toplevels[mention.id] = mention
                mentions.append(mention)
            else:
                # It's a sublabel, set it aside for a second
                sublabels.append(mention)

        # Now match up the sublabels
        base_sets: dict[str, defines.LabelSet] = {}
        for sublabel in sublabels:
            toplevel = toplevels[sublabel.id]

            # Wipe out any toplevel tags (existence of a sublabel implies no toplevel labels)
            if toplevel.id not in base_sets:
                base_sets[toplevel.id] = toplevel.labels
                toplevel.labels = set()

            # Now merge in new labels (preserving other fields like `text` from toplevel entry)
            base_labels = base_sets[toplevel.id]
            for label in sublabel.labels:
                for base_label in base_labels:
                    toplevel.labels.add(
                        defines.Label(base_label.label, sublabel.from_name, label.label)
                    )

        return Annotation(author=author, mentions=mentions)


@dataclasses.dataclass(kw_only=True)
class Note:
    """All of a single note's annotations"""

    note_id: int
    annotations: list[Annotation] = dataclasses.field(default_factory=list)

    # metadata
    docref_mappings: dict[str, str] = dataclasses.field(default_factory=dict)
    encounter_id: str | None = None
    anon_encounter_id: str | None = None

    @staticmethod
    def parse(entry: dict) -> "Note":
        metadata = entry.get("data", {})
        docref_mappings = metadata.get("docref_mappings", {})
        encounter_id = metadata.get("encounter_id") or metadata.get("enc_id")  # old name
        anon_encounter_id = metadata.get("anon_encounter_id") or metadata.get("anon_id")  # old name

        data_keys = set(metadata.keys())
        annotations = [
            Annotation.parse(x, data_keys=data_keys) for x in entry.get("annotations", [])
        ]
        annotations = list(filter(None, annotations))  # parse() returns None if we should skip

        return Note(
            note_id=entry["id"],
            annotations=annotations,
            docref_mappings=docref_mappings,
            encounter_id=encounter_id,
            anon_encounter_id=anon_encounter_id,
        )


class ExportFile:
    """Parse information from Label Studio export files."""

    def __init__(self, path: str):
        """If path is a file, load it. If a folder, merge all export files in it."""
        self._notes = []
        note_hashes = {}

        if os.path.isdir(path):
            filenames = sorted(
                os.path.join(path, name)
                for name in os.listdir(path)
                if name.casefold().endswith(".json")
            )
        else:
            filenames = [path]

        for name in filenames:
            try:
                data = common.read_json(name)
            except Exception as exc:
                print(f"Could not parse '{name}': {exc}", file=sys.stderr)
                continue

            # Confirm it is (very roughly) shaped like an LS export.
            # We are pretty loose, for unit testing's sake.
            # If we end up reading files we shouldn't in the real world, we can add a few more
            # keys to check for.
            if not data or not isinstance(data, list):
                continue
            if not isinstance(data[0], dict):
                continue
            if "id" not in data[0]:
                continue

            # Fold new notes into running list
            new_notes = [Note.parse(x) for x in data]
            for note in new_notes:
                if note.docref_mappings:
                    # Smush the note IDs together to form a stable string
                    ids_hash = hash(tuple(sorted(note.docref_mappings.items())))
                    if old_note := note_hashes.get(ids_hash):
                        self._merge_notes(old_note, note)
                    else:
                        note_hashes[ids_hash] = note
                        self._notes.append(note)
                else:
                    self._notes.append(note)

        if not self._notes:
            errors.exit_for_invalid_project("No Label Studio export data found.")

    @staticmethod
    def _merge_notes(old: Note, new: Note) -> None:
        # Only merge annotations, take all the metadata from the old note
        for new_annot in new.annotations:
            for old_annot in old.annotations:
                if old_annot.author == new_annot.author:
                    old_annot.mentions.extend(new_annot.mentions)
                    break
            else:
                old.annotations.append(new_annot)
                continue

    @property
    def notes(self) -> list[Note]:
        return self._notes



"""Tests for commands/accuracy.py"""

from tests import base


class TestAccuracy(base.TestCase):
    """Test case for the top-level accuracy code"""

    def test_default_output(self):
        stdout = self.run_cli("accuracy", "jill", "jane", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            """Comparing 3 charts (1, 3–4)
Truth: jill
Annotator: jane
Macro F1: 0.833

F1     Sens  Spec   PPV  NPV   Kappa  TP  FN  TN  FP  Label   
0.667  0.75  0.6    0.6  0.75  0.341  3   1   3   2   *       
0.667  0.5   1.0    1.0  0.5   0.4    1   1   1   0   Cough   
1.0    1.0   1.0    1.0  1.0   1.0    2   0   1   0   Fatigue 
-      -     0.333  0.0  1.0   0.0    0   0   1   2   Headache
""",
            stdout,
        )

    def test_csv(self):
        stdout = self.run_cli("accuracy", "--csv", "jill", "jane", path=f"{self.DATA_DIR}/cold")

        self.assertEqual(
            [
                "f1,sens,spec,ppv,npv,kappa,tp,fn,tn,fp,label",
                "0.667,0.75,0.6,0.6,0.75,0.341,3,1,3,2,*",
                "0.667,0.5,1.0,1.0,0.5,0.4,1,1,1,0,Cough",
                "1.0,1.0,1.0,1.0,1.0,1.0,2,0,1,0,Fatigue",
                ",,0.333,0.0,1.0,0.0,0,0,1,2,Headache",
            ],
            stdout.splitlines(),
        )

    def test_verbose(self):
        output = self.run_cli("accuracy", "--verbose", "jill", "jane", path=f"{self.DATA_DIR}/cold")
        self.assertEqual(
            """Comparing 3 charts (1, 3–4)
Truth: jill
Annotator: jane

╭──────────┬──────────┬────────────────╮
│ Chart ID │ Label    │ Classification │
├──────────┼──────────┼────────────────┤
│ 1        │ Cough    │ TP             │
│ 1        │ Fatigue  │ TP             │
│ 1        │ Headache │ FP             │
├──────────┼──────────┼────────────────┤
│ 3        │ Cough    │ TN             │
│ 3        │ Fatigue  │ TN             │
│ 3        │ Headache │ TN             │
├──────────┼──────────┼────────────────┤
│ 4        │ Cough    │ FN             │
│ 4        │ Fatigue  │ TP             │
│ 4        │ Headache │ FP             │
╰──────────┴──────────┴────────────────╯
""",
            output,
        )

    def test_verbose_cvs(self):
        """Verify we can also print verbose results in CSV format"""
        stdout = self.run_cli(
            "accuracy", "--verbose", "--csv", "jill", "jane", path=f"{self.DATA_DIR}/cold"
        )
        self.assertEqual(
            [
                "chart_id,label,classification",
                "1,Cough,TP",
                "1,Fatigue,TP",
                "1,Headache,FP",
                "3,Cough,TN",
                "3,Fatigue,TN",
                "3,Headache,TN",
                "4,Cough,FN",
                "4,Fatigue,TP",
                "4,Headache,FP",
            ],
            stdout.splitlines(),
        )

    def test_bad_truth(self):
        """Verify that we do something suitable for bad arguments"""
        # Truth
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("accuracy", "nope1", "nope2", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Unrecognized annotator 'nope1'\n", stderr.getvalue())

        # Annotator
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("accuracy", "jill", "nope2", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Unrecognized annotator 'nope2'\n", stderr.getvalue())

    def test_same_annotators(self):
        """Verify that we do something suitable for bad arguments"""
        with self.capture_stderr() as stderr:
            with self.assertRaises(SystemExit):
                self.run_cli("accuracy", "jill", "jill", path=f"{self.DATA_DIR}/cold")
        self.assertEqual("Can’t compare the same annotator with themselves.\n", stderr.getvalue())

    def test_sublabels(self):
        output = self.run_cli("accuracy", "alice", "bob", path=f"{self.DATA_DIR}/sublabels")
        self.assertEqual(
            output,
            """Comparing 2 charts (1308–1309)
Truth: alice
Annotator: bob
Macro F1: 1.0

F1     Sens  Spec  PPV   NPV  Kappa  TP  FN  TN  FP  Label                      
0.857  1.0   0.8   0.75  1.0  0.75   3   0   4   1   *                          
1.0    1.0   1.0   1.0   1.0  1.0    1   0   1   0   Deceased → *               
1.0    1.0   1.0   1.0   1.0  1.0    1   0   1   0   Deceased → False           
-      -     0.5   0.0   1.0  0.0    0   0   1   1   Deceased → Datetime → *    
-      0.0   1.0   -     0.5  0.0    0   1   1   0   Deceased → Datetime →      
                                                     11/12/25                   
-      -     0.5   0.0   1.0  0.0    0   0   1   1   Deceased → Datetime →      
                                                     11/13/25                   
1.0    1.0   1.0   1.0   1.0  1.0    1   0   1   0   Fungal → *                 
1.0    1.0   1.0   1.0   1.0  1.0    1   0   1   0   Fungal → Confirmed         
1.0    1.0   1.0   1.0   1.0  1.0    1   0   1   0   Infection → *              
-      -     -     -     -    -      0   0   0   0   Infection → Confirmed      
1.0    1.0   1.0   1.0   1.0  1.0    1   0   1   0   Infection → Suspected      

Ignoring 1 invalid mention
""",
        )



"""Tests for studio.py"""

import json
import tempfile

from chart_review import studio
from tests import base


class TestStudio(base.TestCase):
    """Test some edge cases with Label Studio export parsing"""

    def test_invalid_label_type(self):
        with tempfile.NamedTemporaryFile("wt") as tmpfile:
            json.dump(
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {
                                        "value": {"labels": ["A"]},
                                        "type": "bogus",
                                    }
                                ],
                            },
                        ],
                    },
                ],
                tmpfile,
            )
            tmpfile.flush()
            with self.assertRaisesRegex(ValueError, "Unrecognized Label Studio result type"):
                studio.ExportFile(tmpfile.name)

    def test_merging(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(f"{tmpdir}/one.json", "w", encoding="utf8") as f:
                json.dump(
                    [
                        {
                            "id": 1,
                            "annotations": [
                                {
                                    "completed_by": 1,
                                    "result": [{"value": {"labels": ["LabelA"]}}],
                                },
                                {
                                    "completed_by": 2,
                                    "result": [{"value": {"labels": ["LabelB"]}}],
                                },
                            ],
                            "data": {
                                "docref_mappings": {"A": "anonA", "B": "anonB"},
                            },
                        },
                    ],
                    f,
                )
            with open(f"{tmpdir}/two.json", "w", encoding="utf8") as f:
                json.dump(
                    [
                        {
                            "id": 2,
                            "annotations": [
                                {
                                    "completed_by": 1,
                                    "result": [{"value": {"labels": ["LabelC"]}}],
                                },
                                {
                                    "completed_by": 3,
                                    "result": [{"value": {"labels": ["LabelD"]}}],
                                },
                            ],
                            "data": {
                                "docref_mappings": {"A": "anonA", "B": "anonB"},
                            },
                        },
                    ],
                    f,
                )
            with open(f"{tmpdir}/three.json", "w", encoding="utf8") as f:
                json.dump(
                    [
                        {
                            "id": 3,
                            "annotations": [
                                {
                                    "completed_by": 1,
                                    "result": [{"value": {"labels": ["LabelE"]}}],
                                },
                            ],
                            "data": {
                                # Will not be merged, because this doesn't fully match other notes
                                "docref_mappings": {"A": "anonA"},
                            },
                        },
                    ],
                    f,
                )

            merged = studio.ExportFile(tmpdir)

        self.assertEqual(
            merged.notes,
            [
                studio.Note.parse(
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {"value": {"labels": ["LabelA"]}},
                                    {"value": {"labels": ["LabelC"]}},
                                ],
                            },
                            {
                                "completed_by": 2,
                                "result": [{"value": {"labels": ["LabelB"]}}],
                            },
                            {
                                "completed_by": 3,
                                "result": [{"value": {"labels": ["LabelD"]}}],
                            },
                        ],
                        "data": {
                            "docref_mappings": {"A": "anonA", "B": "anonB"},
                        },
                    }
                ),
                studio.Note.parse(
                    {
                        "id": 3,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [{"value": {"labels": ["LabelE"]}}],
                            },
                        ],
                        "data": {
                            "docref_mappings": {"A": "anonA"},
                        },
                    }
                ),
            ],
        )

    def test_bad_file(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(f"{tmpdir}/invalid.json", "w", encoding="utf8") as f:
                f.write('[{"id": ]')
            with open(f"{tmpdir}/strings.json", "w", encoding="utf8") as f:
                f.write('["string list"]')
            with open(f"{tmpdir}/no-id.json", "w", encoding="utf8") as f:
                f.write('[{"random": "yup"}]')
            with open(f"{tmpdir}/valid.json", "w", encoding="utf8") as f:
                f.write('[{"id": 1}]')

            self.assertEqual(
                studio.ExportFile(tmpdir).notes,
                [studio.Note.parse({"id": 1})],
            )

    def test_supported_types(self):
        with tempfile.NamedTemporaryFile("wt") as tmpfile:
            json.dump(
                [
                    {
                        "id": 1,
                        "annotations": [
                            {
                                "completed_by": 1,
                                "result": [
                                    {
                                        "value": {"choices": ["Choice"]},
                                        "type": "choices",
                                    },
                                    {
                                        "value": {"datetime": "2021-04-22"},
                                        "type": "datetime",
                                    },
                                    {
                                        "value": {"datetime": "2021-04-22T10:20:30Z"},
                                        "type": "datetime",
                                    },
                                    {
                                        "value": {"labels": ["Label"]},
                                        "type": "labels",
                                    },
                                    {
                                        "value": {"text": ["Text"]},
                                        "type": "textarea",
                                    },
                                ],
                            },
                        ],
                    },
                ],
                tmpfile,
            )
            tmpfile.flush()
            export = studio.ExportFile(tmpfile.name)

        self.assertEqual(
            [x.labels for x in export.notes[0].annotations[0].mentions],
            [
                base.labels(["Choice"]),
                base.labels(["2021-04-22"]),
                base.labels(["2021-04-22T10:20:30Z"]),
                base.labels(["Label"]),
                base.labels(["Text"]),
            ],
        )



"""Tests for agree.py"""

import math

import ddt

from chart_review import agree, defines
from tests import base


@ddt.ddt
class TestAgreement(base.TestCase):
    """Test case for basic agreement logic"""

    @ddt.data(
        (
            "alice",
            "bob",
            None,
            {
                "FN": [(1, "Cough", "")],
                "FP": [(1, "Headache", ""), (2, "Cough", "")],
                "TN": [(1, "Fever", ""), (2, "Headache", "")],
                "TP": [(2, "Fever", "")],
            },
        ),
        (
            "bob",
            "alice",
            {},
            {
                "FN": [(1, "Headache", ""), (2, "Cough", "")],
                "FP": [(1, "Cough", "")],
                "TN": [(1, "Fever", ""), (2, "Headache", "")],
                "TP": [(2, "Fever", "")],
            },
        ),
        (
            "alice",
            "bob",
            base.labels(["Cough"]),
            {
                "FN": [(1, "Cough", "")],
                "FP": [(2, "Cough", "")],
                "TN": [],
                "TP": [],
            },
        ),
    )
    @ddt.unpack
    def test_confusion_matrix_counts(self, truth, annotator, labels, expected_matrix):
        """Verify that we can make a simple confusion matrix."""
        annotations = defines.ProjectAnnotations(
            labels={"Cough", "Fever", "Headache"},
            mentions={
                "alice": {1: base.labels({"Cough"}), 2: base.labels({"Fever"})},
                "bob": {1: base.labels({"Headache"}), 2: base.labels({"Cough", "Fever"})},
            },
        )
        notes = [1, 2]

        matrix = agree.confusion_matrix(annotations, truth, annotator, notes, labels=labels)
        self.assertEqual(expected_matrix, matrix)

    def test_confusion_matrix_for_sublabels(self):
        annotations = defines.ProjectAnnotations(
            labels=base.labels({"Top|Sub|A", "Top|Sub|B", "Top|Sub|C"}),
            mentions={
                "alice": {
                    1: base.labels({"Top|Sub|A"}),  # TP
                    2: base.labels({"Top|Sub|A"}),  # FP
                    3: base.labels({"Top|Sub|A", "Top|Sub|B"}),  # TP
                    4: base.labels({"Top|Sub|A", "Top|Sub|B"}),  # FP
                    5: set(),  # TN
                    6: base.labels({"Top|Sub|A"}),  # FN
                    7: set(),  # FP
                },
                "bob": {
                    1: base.labels({"Top|Sub|A"}),  # TP
                    2: base.labels({"Top|Sub|C"}),  # FP
                    3: base.labels({"Top|Sub|A", "Top|Sub|B"}),  # TP
                    4: base.labels({"Top|Sub|C", "Top|Sub|B"}),  # FP
                    5: set(),  # TN
                    6: set(),  # FN
                    7: base.labels({"Top|Sub|A"}),  # FP
                },
            },
        )
        notes = [1, 2, 3, 4, 5, 6, 7]

        matrix = agree.confusion_matrix(annotations, "alice", "bob", notes)
        self.assertEqual(
            matrix,
            {
                "FN": [(6, "Top", "Sub")],
                "FP": [(2, "Top", "Sub"), (4, "Top", "Sub"), (7, "Top", "Sub")],
                "TN": [(5, "Top", "Sub")],
                "TP": [(1, "Top", "Sub"), (3, "Top", "Sub")],
            },
        )

    @ddt.data(
        # Examples pulled from https://en.wikipedia.org/wiki/Cohen's_kappa#Examples
        (
            {
                "FN": [{x: "Label"} for x in range(5)],
                "FP": [{x: "Label"} for x in range(10)],
                "TN": [{x: "Label"} for x in range(15)],
                "TP": [{x: "Label"} for x in range(20)],
            },
            0.4,
        ),
        (
            {
                "FN": [{x: "Label"} for x in range(15)],
                "FP": [{x: "Label"} for x in range(25)],
                "TN": [{x: "Label"} for x in range(15)],
                "TP": [{x: "Label"} for x in range(45)],
            },
            0.1304,
        ),
        (
            {
                "FN": [{x: "Label"} for x in range(35)],
                "FP": [{x: "Label"} for x in range(5)],
                "TN": [{x: "Label"} for x in range(35)],
                "TP": [{x: "Label"} for x in range(25)],
            },
            0.2593,
        ),
        # This example is from table 2 in https://pubmed.ncbi.nlm.nih.gov/12474424/
        (
            {
                "FN": [{x: "Label"} for x in range(6)],
                "FP": [{x: "Label"} for x in range(9)],
                "TN": [{x: "Label"} for x in range(26)],
                "TP": [{x: "Label"} for x in range(15)],
            },
            0.4444,
        ),
    )
    @ddt.unpack
    def test_kappa_score(self, matrix, expected_kappa):
        """Verify that we can score a matrix for kappa."""
        kappa = round(agree.score_kappa(matrix), 4)
        self.assertEqual(expected_kappa, kappa)

    @ddt.data(
        # Example of incredibly unbalanced data, as a regression test
        (
            {
                "FN": [],
                "FP": [],
                "TN": [{x: "Label"} for x in range(90)],
                "TP": [],
            },
        ),
        (
            {
                "FN": [],
                "FP": [],
                "TN": [],
                "TP": [{x: "Label"} for x in range(90)],
            },
        ),
    )
    @ddt.unpack
    def test_unbalanced_kappa(self, matrix):
        """Verify that kappa will handle unbalanced NaN cases."""
        kappa = round(agree.score_kappa(matrix), 4)
        self.assertEqual(math.isnan(kappa), True)
